from decimal import Decimal

from lxml import html

from fastexchange.currency import SUPPORTED_CTS
from fastexchange.exceptions import ConverterNotMapped
from fastexchange.http import validate_response

from .base_client import BaseConverter, DiscriminatedCurrency, ToCurrency
from .mappings import (
    EURToUSD,
    FromToExchange,
    USDToEur,
)


def exchange(from_: DiscriminatedCurrency, to: type[ToCurrency], rate: Decimal) -> ToCurrency:
    to_type: SUPPORTED_CTS = to.__annotations__["c_type"].__args__[0]
    result = from_.convert_to(to_type, rate)
    return result  # type: ignore


class CurrencyMeUkClient(BaseConverter):
    mapping: dict[FromToExchange, str] = {
        USDToEur: "/usd/eur",
        EURToUSD: "/eur/usd",
    }
    base_url = "https://www.currency.me.uk/convert"

    async def exchange(self, from_: DiscriminatedCurrency, to: type[ToCurrency]) -> ToCurrency:
        query_str = self.mapping.get(FromToExchange(from_=type(from_), to=to), None)
        if query_str is None:
            raise ConverterNotMapped(
                f"Unsupported conversion from {type(from_)} to {to} by {self.__class__.__name__}"
            )
        resp = await self.client.get(f"{self.base_url}{query_str}")
        with validate_response(resp, 200):
            tree = html.fromstring(resp.content)
            exchange_ratios = tree.xpath('//input[@id="answer"]/@value')
            assert len(exchange_ratios) > 0
            exchange_rate = Decimal(exchange_ratios[0])
        return exchange(from_, to, exchange_rate)



import typing
from contextlib import contextmanager
from functools import lru_cache
from http.cookiejar import CookieJar

from httpx import AsyncClient
from httpx._client import USE_CLIENT_DEFAULT, UseClientDefault
from httpx._config import Timeout
from httpx._models import Request, Response
from httpx._types import (
    AuthTypes,
    CookieTypes,
    HeaderTypes,
    QueryParamTypes,
    RequestContent,
    RequestData,
    RequestExtensions,
    RequestFiles,
    TimeoutTypes,
    URLTypes,
)
from typing_extensions import Self


class NullCookieJar(CookieJar):
    """A CookieJar that does not support setting cookie"""

    def extract_cookies(self, *args, **kwargs):
        """For extracting and saving cookies.  This implementation does nothing"""
        pass

    def set_cookie(self, *args, **kwargs):
        """Normally for setting a cookie.  This implementation does nothing"""
        pass


class BaseClient:
    """A wrapper around httpx.AsyncClient
    This client will be used as singleton to get best performance; quote from httpx doc:
    In order to get the most benefit from connection pooling, make sure you're not instantiating multiple client instances
    - for example by using async with inside a "hot loop".
    This can be achieved either by having a single scoped client that's passed throughout wherever
    it's needed, or by having a single global client instance.
    This client uses `NullCookieJar` to make sute the httpx singleton client doesn't have a global mutable state.
    E.x: 2 request is done to same origin, request A set session id as "X", request B set session id as "Y". but in fact,
        client will end up having only one of them set (for instance X), and the rest request would be send with session id X.
        This would result in getting the client easily detected.
    But with this implementation, client isolate request and response from its global state.
    If you need to set cookie on next request, you should do it explicitly.
    """

    def __init__(self, *args, **kwargs) -> None:
        self._client = AsyncClient(*args, **kwargs)
        self._client.cookies = NullCookieJar()

    async def send(self, *args, **kwargs) -> Response:
        return await self._client.send(*args, **kwargs)

    async def __aenter__(self) -> Self:
        await self._client.__aenter__()
        return self

    async def __aexit__(self, *args, **kwargs) -> None:
        await self._client.__aexit__(*args, **kwargs)

    async def aclose(self) -> None:
        await self._client.aclose()

    def build_request(
        self,
        method: str,
        url: URLTypes,
        *,
        content: typing.Optional[RequestContent] = None,
        data: typing.Optional[RequestData] = None,
        files: typing.Optional[RequestFiles] = None,
        json: typing.Optional[typing.Any] = None,
        params: typing.Optional[QueryParamTypes] = None,
        headers: typing.Optional[HeaderTypes] = None,
        cookies: typing.Optional[CookieTypes] = None,
        timeout: typing.Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
        extensions: typing.Optional[RequestExtensions] = None,
    ) -> Request:
        extensions = {} if extensions is None else extensions
        if "timeout" not in extensions:
            timeout = (
                self._client.timeout if isinstance(timeout, UseClientDefault) else Timeout(timeout)
            )
            extensions = dict(**extensions, timeout=timeout.as_dict())
        return Request(
            method,
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            extensions=extensions,
        )

    async def request(
        self,
        method: str,
        url: URLTypes,
        *,
        content: typing.Optional[RequestContent] = None,
        data: typing.Optional[RequestData] = None,
        files: typing.Optional[RequestFiles] = None,
        json: typing.Optional[typing.Any] = None,
        params: typing.Optional[QueryParamTypes] = None,
        headers: typing.Optional[HeaderTypes] = None,
        cookies: typing.Optional[CookieTypes] = None,
        auth: typing.Union[AuthTypes, UseClientDefault, None] = USE_CLIENT_DEFAULT,
        follow_redirects: typing.Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
        timeout: typing.Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
        extensions: typing.Optional[RequestExtensions] = None,
    ) -> Response:
        request = self.build_request(
            method=method,
            url=url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            timeout=timeout,
            extensions=extensions,
        )
        return await self.send(request, auth=auth, follow_redirects=follow_redirects)

    async def get(
        self,
        url: URLTypes,
        *,
        params: typing.Optional[QueryParamTypes] = None,
        headers: typing.Optional[HeaderTypes] = None,
        cookies: typing.Optional[CookieTypes] = None,
        auth: typing.Union[AuthTypes, UseClientDefault, None] = USE_CLIENT_DEFAULT,
        follow_redirects: typing.Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
        timeout: typing.Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
        extensions: typing.Optional[RequestExtensions] = None,
    ) -> Response:
        return await self.request(
            "GET",
            url,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )

    async def post(
        self,
        url: URLTypes,
        *,
        content: typing.Optional[RequestContent] = None,
        data: typing.Optional[RequestData] = None,
        files: typing.Optional[RequestFiles] = None,
        json: typing.Optional[typing.Any] = None,
        params: typing.Optional[QueryParamTypes] = None,
        headers: typing.Optional[HeaderTypes] = None,
        cookies: typing.Optional[CookieTypes] = None,
        auth: typing.Union[AuthTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
        follow_redirects: typing.Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
        timeout: typing.Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
        extensions: typing.Optional[RequestExtensions] = None,
    ) -> Response:
        return await self.request(
            "POST",
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )

    async def put(
        self,
        url: URLTypes,
        *,
        content: typing.Optional[RequestContent] = None,
        data: typing.Optional[RequestData] = None,
        files: typing.Optional[RequestFiles] = None,
        json: typing.Optional[typing.Any] = None,
        params: typing.Optional[QueryParamTypes] = None,
        headers: typing.Optional[HeaderTypes] = None,
        cookies: typing.Optional[CookieTypes] = None,
        auth: typing.Union[AuthTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
        follow_redirects: typing.Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
        timeout: typing.Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
        extensions: typing.Optional[RequestExtensions] = None,
    ) -> Response:
        """
        Send a `PUT` request.
        **Parameters**: See `httpx.request`.
        """
        return await self.request(
            "PUT",
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )

    async def patch(
        self,
        url: URLTypes,
        *,
        content: typing.Optional[RequestContent] = None,
        data: typing.Optional[RequestData] = None,
        files: typing.Optional[RequestFiles] = None,
        json: typing.Optional[typing.Any] = None,
        params: typing.Optional[QueryParamTypes] = None,
        headers: typing.Optional[HeaderTypes] = None,
        cookies: typing.Optional[CookieTypes] = None,
        auth: typing.Union[AuthTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
        follow_redirects: typing.Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
        timeout: typing.Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
        extensions: typing.Optional[RequestExtensions] = None,
    ) -> Response:
        return await self.request(
            "PATCH",
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )

    async def delete(
        self,
        url: URLTypes,
        *,
        params: typing.Optional[QueryParamTypes] = None,
        headers: typing.Optional[HeaderTypes] = None,
        cookies: typing.Optional[CookieTypes] = None,
        auth: typing.Union[AuthTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
        follow_redirects: typing.Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
        timeout: typing.Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
        extensions: typing.Optional[RequestExtensions] = None,
    ) -> Response:
        """
        Send a `DELETE` request.
        **Parameters**: See `httpx.request`.
        """
        return await self.request(
            "DELETE",
            url,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )

    async def options(
        self,
        url: URLTypes,
        *,
        params: typing.Optional[QueryParamTypes] = None,
        headers: typing.Optional[HeaderTypes] = None,
        cookies: typing.Optional[CookieTypes] = None,
        auth: typing.Union[AuthTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
        follow_redirects: typing.Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
        timeout: typing.Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
        extensions: typing.Optional[RequestExtensions] = None,
    ) -> Response:
        """
        Send an `OPTIONS` request.
        **Parameters**: See `httpx.request`.
        """
        return await self.request(
            "OPTIONS",
            url,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )

    async def head(
        self,
        url: URLTypes,
        *,
        params: typing.Optional[QueryParamTypes] = None,
        headers: typing.Optional[HeaderTypes] = None,
        cookies: typing.Optional[CookieTypes] = None,
        auth: typing.Union[AuthTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
        follow_redirects: typing.Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
        timeout: typing.Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
        extensions: typing.Optional[RequestExtensions] = None,
    ) -> Response:
        """
        Send a `HEAD` request.
        **Parameters**: See `httpx.request`.
        """
        return await self.request(
            "HEAD",
            url,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )


@contextmanager
def validate_response(resp: Response, expected_status_code: int):
    if resp.status_code != expected_status_code:
        raise RequestFailed(resp)

    try:
        yield
    except Exception:
        raise RequestFailed(resp)


class RequestFailed(Exception):
    def __init__(self, resp: Response):
        message = (
            f"Response: {resp.text}\nRequest: url={resp.request.url} headers={resp.request.headers}"
        )
        super().__init__(message)


@lru_cache
def get_client():
    return BaseClient()



from typing import NamedTuple

from ..currency import DiscriminatedCurrency, EURCurrency, USDCurrency


class FromToExchange(NamedTuple):
    from_: type[DiscriminatedCurrency]
    to: type[DiscriminatedCurrency]


USDToEur = FromToExchange(from_=USDCurrency, to=EURCurrency)
EURToUSD = FromToExchange(from_=EURCurrency, to=USDCurrency)



from __future__ import annotations

from datetime import datetime
from decimal import Decimal
from typing import Annotated, Union

from pydantic import BaseModel, BeforeValidator, Field

from fastexchange.utils import StrEnum

from ..currency import USDCurrency
from .types import UsdtTrcToken

UnitMsTime = Annotated[
    datetime,
    BeforeValidator(lambda v: datetime.fromtimestamp(v / 1000) if isinstance(v, int) else v),
]


class EventTypes(StrEnum):
    TRANSFER = "Transfer"


class Status(StrEnum):
    SUCCESS = "SUCCESS"


class Network(StrEnum):
    TRC20 = "trc20"


class TokenInfo(BaseModel):
    tokenId: str
    tokenAbbr: str
    tokenName: str
    tokenDecimal: int
    tokenCanShow: int
    tokenType: str
    tokenLogo: str
    tokenLevel: str
    issuerAddr: str
    vip: bool


class TokenTransfer(BaseModel):
    transaction_id: str
    status: int
    block_ts: UnitMsTime
    from_address: str
    to_address: str
    block: int
    contract_address: str
    quant: Decimal
    approval_amount: int

    confirmed: bool

    revert: bool

    fromAddressIsContract: bool
    toAddressIsContract: bool
    riskTransaction: bool

    # allow to parse others
    event_type: Union[EventTypes, str] = Field(union_mode="left_to_right")
    contract_type: Union[Network, str] = Field(union_mode="left_to_right")
    finalResult: Union[Status, str] = Field(union_mode="left_to_right")
    contractRet: Union[Status, str] = Field(union_mode="left_to_right")

    def to_usd(self, current_rate: Decimal = Decimal(1)) -> USDCurrency:
        """We assume that 1 usdt == 1 usd, but in case this wasn't true, this method should get the current rate"""
        token = self.to_token()
        return USDCurrency(val=token.val / current_rate)

    def to_token(self) -> UsdtTrcToken:
        val = self.quant / 1000000
        return UsdtTrcToken(val=val)


class Transfers(BaseModel):
    total: int
    rangeTotal: int
    token_transfers: list[TokenTransfer]



from abc import ABC
from decimal import Decimal
from enum import auto
from typing import Annotated, Literal, Union

from pydantic import BaseModel, Field, TypeAdapter

from fastexchange.utils import StrEnum


class CryptoToken(StrEnum):
    USDT_TRC20 = auto()
    USDT_ETH = auto()


class BaseCryptoToken(BaseModel, ABC):
    t_type: CryptoToken
    """Token Type"""
    val: Decimal
    """Token Value"""


class UsdtTrcToken(BaseCryptoToken):
    t_type: Literal[CryptoToken.USDT_TRC20] = CryptoToken.USDT_TRC20


class UsdtEthToken(BaseCryptoToken):
    t_type: Literal[CryptoToken.USDT_ETH] = CryptoToken.USDT_ETH


DiscriminatedCryptoToken = Annotated[
    Union[UsdtTrcToken, UsdtEthToken], Field(discriminator="t_type")
]
AnyCryptoToken = TypeAdapter(DiscriminatedCryptoToken)



from .converter.clients import CurrencyMeUkClient
from .crypto.types import UsdtTrcToken
from .crypto.usdt import Transfers, Trc20Gateway
from .currency import AnyCurrency, BaseCurrency, EURCurrency, USDCurrency
from .exceptions import CantCompareException, ConverterNotMapped

__all__ = [
    "CurrencyMeUkClient",
    "BaseCurrency",
    "USDCurrency",
    "EURCurrency",
    "AnyCurrency",
    "Trc20Gateway",
    "Transfers",
    "ConverterNotMapped",
    "CantCompareException",
    "UsdtTrcToken",
]



from enum import Enum


class StrEnum(str, Enum):
    """
    StrEnum subclasses that create variants using `auto()` will have values equal to their names

    Enums inheriting from this class that set values using `enum.auto()` will have variant values
        equal to their names
    """

    @staticmethod
    def _generate_next_value_(name: str, start: int, count: int, last_values: list[str]) -> str:
        """
        Uses the name as the automatic value, rather than an integer
        See https://docs.python.org/3/library/enum.html#using-automatic-values for reference
        """
        return name

    def __str__(self) -> str:
        return str(self.value)



from abc import ABC, abstractmethod
from typing import Union

from typing_extensions import TypeVar

from fastexchange.converter.mappings import FromToExchange
from fastexchange.http import BaseClient, get_client

from ..currency import DiscriminatedCurrency

ToCurrency = TypeVar("ToCurrency", bound=DiscriminatedCurrency)


class BaseConverter(ABC):
    mapping: dict[FromToExchange, str]

    def __init__(self, client: Union[BaseClient, None] = None):
        self.client = client or get_client()

    def is_supported(self, from_: type[DiscriminatedCurrency], to: type[ToCurrency]) -> bool:
        return FromToExchange(from_=from_, to=to) in self.mapping

    @abstractmethod
    async def exchange(self, _from: DiscriminatedCurrency, to: type[ToCurrency]) -> ToCurrency: ...



class BaseFinancialException(Exception):
    ...


class CantCompareException(BaseFinancialException):
    """Cannot compare two different currencies"""


class ConverterNotMapped(BaseFinancialException):
    """Converter Client does not support requested spefiec exchange"""



from abc import ABC
from decimal import Decimal
from enum import auto
from typing import Annotated, Generic, Literal, TypeVar, Union

from pydantic import BaseModel, Field, TypeAdapter
from typing_extensions import TypeAlias

from fastexchange.utils import StrEnum

from .exceptions import CantCompareException


class CurrencyEnum(StrEnum):
    USD = auto()
    EURO = auto()


SUPPORTED_CTS: TypeAlias = Literal[CurrencyEnum.EURO, CurrencyEnum.USD]

CT = TypeVar("CT", bound=SUPPORTED_CTS)
OtherCT = TypeVar("OtherCT", bound=SUPPORTED_CTS)


class BaseCurrency(BaseModel, Generic[CT], ABC):
    c_type: CT
    """Currency type"""
    val: Decimal
    """Currency value"""

    def convert_to(self, other_currency: OtherCT, rate: Decimal) -> "BaseCurrency[OtherCT]":
        return AnyCurrency.validate_python(
            {"c_type": other_currency, "val": self.val * rate}  # type: ignore
        )

    def check_if_self(self, other: "BaseCurrency[CT]"):
        # type system avoid this state, but still ehh
        if self.c_type != other.c_type:
            raise CantCompareException(f"Cannot do action {self.c_type} with {other.c_type}")

    def __eq__(self, other: "BaseCurrency[CT]") -> bool:
        self.check_if_self(other)
        return self.val == other.val

    def __lt__(self, other: "BaseCurrency[CT]") -> bool:
        self.check_if_self(other)
        return self.val < other.val

    def __le__(self, other: "BaseCurrency[CT]") -> bool:
        self.check_if_self(other)
        return self < other or self == other

    def __gt__(self, other: "BaseCurrency[CT]") -> bool:
        self.check_if_self(other)
        return not (self <= other)

    def __ge__(self, other: "BaseCurrency[CT]") -> bool:
        self.check_if_self(other)
        return not (self < other)

    def __add__(self, other: "BaseCurrency[CT]") -> "BaseCurrency[CT]":
        self.check_if_self(other)
        return self.__class__(c_type=self.c_type, val=self.val + other.val)

    def __sub__(self, other: "BaseCurrency[CT]") -> "BaseCurrency[CT]":
        self.check_if_self(other)
        return self.__class__(c_type=self.c_type, val=self.val - other.val)

    def __hash__(self) -> int:
        return hash((self.c_type, self.val))


class USDCurrency(BaseCurrency[Literal[CurrencyEnum.USD]]):
    c_type: Literal[CurrencyEnum.USD] = CurrencyEnum.USD


class EURCurrency(BaseCurrency[Literal[CurrencyEnum.EURO]]):
    c_type: Literal[CurrencyEnum.EURO] = CurrencyEnum.EURO


DiscriminatedCurrency = Annotated[Union[USDCurrency, EURCurrency], Field(discriminator="c_type")]
AnyCurrency = TypeAdapter(DiscriminatedCurrency)



from typing import Union

from fastexchange.http import BaseClient, get_client, validate_response

from .schema import Transfers


class Trc20Gateway:
    def __init__(self, client: Union[BaseClient, None] = None):
        self.client = client or get_client()
        self.base_url = "https://apilist.tronscanapi.com/api/token_trc20"

    async def check_transaction(self, wallet: str) -> Transfers:
        resp = await self.client.get(
            f"{self.base_url}/transfers?limit=50&start=0&sort=-timestamp&count=true&relatedAddress={wallet}",
        )
        with validate_response(resp, 200):
            return Transfers.model_validate(resp.json())
